// --- [ led_pwm12bit.c ] -----------------------------------------------------
//
// tab = 3 - 28.12.09
//
// Ansteuerung von bis zu 8 LED's mit 12Bit-PWM bei 76 Hz.
// Die gewnschte Helligkeit wird mit einem 1-Byte-Wert pro LED bergeben.
// Aus einer Tabelle wird spter der zugehrige Compare-Wert aus dem
// 12-Bit Raum ausgewhlt.
// Die Schaltzeitpunkte werden vorsortiert und in zeitlicher Reihenfolge
// als Compare-Wert an den Timer1 rausgeschrieben
// Eine Interrupt-Routine schreibt dann den nchsten Comparewert
// sowie den bereits vorbereiteten Bit-Status auf den LED-Port.
//
// Neue Helligkeitswerte knnen via TWI bergeben werden, sie werden
// dann einen PWM-Takte spter bernommen.
// Alternativ kann ein bergang vom aktuellen LED-Status zu einem
// neu angegebenen veranlasst werden - die Dauer des bergangs bzw. die
// Anzahl der Zwischenschritte ist dabei als 1. Parameter anzugeben.
// Zustzlich kann noch ein 'Verzgerungsfaktor' eingebaut werden:
// normalerweise wird nach jedem PWM-Takt ein neuer Helligkeitswert berechnet,
// bei 76Hz dauert ein bergang (Parameter sei 255) 255/76 also ca. 3 Sekunden
// Wenn pwm_wait = n, dann wird erst nach n-Takten eine neue Helligkeit gerechnet,
// ein Helligkeitswechsel kann sich dann maximal 255/76 * 255 = 14 Minuten hinziehen ...
// In der Zwischenzeit kann der TWI-Slave abgefragt werden:
// - liefert ein TRUE (> 0), dann ist das Gert beschftigt - Daten werden zwar
// angenommen, aber nicht bearbeitet.
// Nur wenn der TWI-Slave ein FALSE (0) liefert, ist er bereit.
//
// ----------------------------------------------------------------------------
#define EXTERN

#include <avr/io.h>
#include <inttypes.h>
#include <stdlib.h>
#include <string.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "TWI_led_slave.h"
#include "global.h"
#include "rs232rx.h"

#include "ocra_tab.h"							// die PWM-Wert als PROGMEM ocra_tab[256]

uint8_t  ausgangswert[LEDs];					// die unsortierten, aktuellen Helligkeitswerte
														// bentigt beim berblenden
uint8_t led_value[LEDs + 2];					// bermittelter 1-Byte-Wert fr Helligkeit
uint8_t led_bitcode[LEDs + 2];				// das zu schaltende Port-Bit fr den zugeh. Kanal
uint8_t pwm_buffer[64];							// hier werden die vorberechneten Werte fr OCR1A und Portpins abgelegt
uint8_t portbit[8];								// Portbits zum Schalten der einzelnen LEDs
uint8_t pwm_wait = 0;                 		// verlangsamt den bergang durch Wartephasen
volatile uint8_t pwm_update;					// zeigt an, ob genderte PWM-Werte vorliegen
volatile uint8_t pwm_counter;					// zhlt PWM-Takte fr Verzgerung
//uint8_t isr_start_idx;						// neuer Startindex fr die ICR1-ISR
#define isr_start_idx TCNT0					// hier wird ein Register als Speicherplatz genutzt
														// spart ca. 1 Takt ein
//uint8_t isr_idx;								// laufender Index der ICR1-ISR
#define isr_idx GPIOR0							// hier wird ein Register als Speicherplatz genutzt
														// spart ca. 3 Takte ein

void fade(uint8_t schrittzahl); 				// Berechnet die Zwischenschritte zwischen zwei Helligkeiten
void prep_values(void);							// Bereitet die Helligkeitswerte fr die PWM-Ausgabe auf
void random_mode(void);							// Zufallsmodus mit stndig neuen, zuflligen Zielwerten


// ---------------------------------------------------------------------------
void delay_10ms(uint8_t s)
{
while (s--) _delay_ms(10);
}

// ----------------------------------------------------------------------------
// Schreibt Zufallszahlen in die Helligkeitswert
// Sinnvoll ist hier, einen Wert von pwm_wait > 0 zu setzen, um die bergnge
// zu verlangsamen ..

void random_mode(void)
{
uint8_t i;

while (!(TWI_ready()))							// Solange wie kein neuer Befehl eingegangen ist
	{
	for (i = 0; i < LEDs; i++) 
		{
		led_value[i] = (rand() | 128); 		// Zufallswert, aber mindestens 128
		}

	fade(255);										// maximale Anzahl von Zwischenschritten
	
	delay_10ms(100);								// kleine Pause einlegen
	}
}


// ----------------------------------------------------------------------------
int main(void)
{
uint8_t i;

// --- Ausgnge fr die LEDs setzen in Ausgangsposition
#ifdef HIGHSIDE
LED_PORT = LED_PINS;
#else
LED_PORT = 0;
#endif

LED_DDR  = LED_PINS + (1<<PD1);


// Position im Kommando 						// Zuordnung der LEDs zu den Eingabebytes
			    // Ausgabepin dafr
portbit[0] = LED7;								// der 1. Wert wird dem Port zugewiesen, der in LED7 definiert ist
portbit[1] = LED6;								// der 2. Wert wird dem Port zugewiesen, der in LED6 definiert ist
portbit[2] = LED5;								// der 3. Wert wird dem Port zugewiesen, der in LED5 definiert ist								
portbit[3] = LED4;								// der 4. Wert wird dem Port zugewiesen, der in LED4 definiert ist
portbit[4] = LED3;								// und so weiter
portbit[5] = LED2;
portbit[6] = LED1;
portbit[7] = 0;									// fr RS232-Eingabe muss der Pin freibleiben

TWI_init();											// TWI initialisieren
rs232_init();										// 9600 Baud

// --- 16-Bit Timer1 Konfigurieren
														// Mode 12 CTC, Prescale 64
TCCR1B = ( 1<<WGM12 | 1<<WGM13 | 0<<CS12 | 1<<CS11 | 1<<CS10 );
ICR1   = PWM_MAX;									// oberes Ende des Zhlers

TIMSK1 = (1<<ICIE1 | 1<< OCIE1A );			// Interrupts freigeben
sei();

// --- Als Ausgangsposition die LED-Helligkeitswerte als eine Rampe setzen
for (i = 0; i < LEDs; i++)
	{
	led_value[i] = led_value[i] = 143 + i * 16; // alternativ alles auf 0/255 setzen
	}
	
prep_values();										// Daten aufbereiten fr den ersten PWM-Start


// --- Hauptprogrammschleife --------------------------------------------------
while(1)
	{
	
	while (!(TWI_ready()));						// warten bis neue Daten angekommen sind

														// das erste Byte dient als Steuerzeichen:
														// 0 -> festen Helligkeitswert setzen
														// 1 -> Variable pwm_wait verndern
														// 2 -> ZUfallsmodus
														// x -> Schrittzahl fr eine berblendung
														
	if (twi_buffer[0] == 1)						// der Variablen 'pwm_wait' einen Wert zuweisen
		{
		pwm_wait = twi_buffer[1];				// den Wert aus twi_buffer[1] kopieren
		TWI_clear_received();					// TWI-Buffer und Index lschen
		continue;
		}

	else if (twi_buffer[0] == 2)				// den Automodus aufrufen
		{
		TWI_clear_received();					// TWI-Buffer und Index lschen
		random_mode();
		continue;
		}

	for (i = 1; i < (LEDs + 1); i++)			// fr alle anderen Flle
		{
		led_value[i - 1] = twi_buffer[i];	// die LED-Helligkeits-Werte umkopieren
		}												// ohne das Steuerbyte

														// Byte 0 als Steuerbyte auswerten
	if (twi_buffer[0] == 0)						// wenn der Parameter == 0, dann
		{
		prep_values();								// die Wert aufbereiten und direkt ausgeben
		}

	else												//  > 2 -> dann ist ein bergang gewnscht
		{

		fade(twi_buffer[0]);						// die Schrittzahl bergeben
		pwm_busy = FALSE;
		}
	TWI_clear_received();						// TWI-Buffer und Index lschen
	}
}

// ----------------------------------------------------------------------------
// wird bei einem CompareA-Match aufgerufen, 
// es wird der nchste Comparewert geschrieben
// und der vorberechnete Portstatus gesetzt.
ISR(TIMER1_COMPA_vect)
{
register uint8_t i;
i = isr_idx;

OCR1AH   = pwm_buffer[i++];
OCR1AL   = pwm_buffer[i++];
LED_PORT = pwm_buffer[i++];

isr_idx = i;
}

// ----------------------------------------------------------------------------
// wird aufgerufen, wenn der Counter den Maximalwert erreicht hat,
// setzt den PWM-Index wieder auf 0 und schaltet die LED's ein,
// sofern nicht einzelne LEDs abgeschaltet sind.
ISR(TIMER1_CAPT_vect)
{
// welche LED-Pins zu setzen sind, ist bereits im Voraus berechnet worden,
// ebenso der Wert fr OCR1A (fr den nchsten Statuswechsel der LEDPins).
register uint8_t i;
i = isr_idx;

if (pwm_update)									// wenn genderte Werte vorhanden sind
	{
	pwm_update = FALSE;							// Flag zurcksetzen
	}

i = isr_start_idx;								// Index zurcksetzen, entweder auf neuen Startwert
														// der von sort_bytes() geschrieben wurde oder auf
														// den letzten Wert zurcksetzen
OCR1AH   = pwm_buffer[i++];					// Compare-Wert schreiben
OCR1AL   = pwm_buffer[i++];
LED_PORT = pwm_buffer[i++];					// Portstatus schreiben
isr_idx = i;										// den Index wieder im Register isr_idx speichern
pwm_counter--;										// nur zur Verzgerung beim Faden erforderlich
}

// ----------------------------------------------------------------------------
// Blendet von Ausgangswert zum Zielwert in der angegebenen Schrittzahl ber,
// verwendet die globalen Variablen:
// - ausgangswert[],
// - led_value[],
// - pwm_wait.
// Zur Verbesserung der Genauigkeit/Verringerung von Rundungsfehlern
// werden alle Zwischenwerte mit 128 multipliziert.
void fade(uint8_t schrittzahl)
{
uint8_t zielwert[LEDs];                 	// den Zielwert der Helligkeit speichern
int16_t zwischenwert[LEDs];               // Wert * 128 fr Zwischenschritt
int16_t delta[LEDs];                      // Differenz * 128 zwischen 2 Schritten
uint8_t i;

pwm_busy = TRUE;									// via TWI kann abgefragt werden, ob ein bergang in Arbeit ist
														// den Zielwert zwischenspeichern
for (i = 0; i < LEDs; i++) zielwert[i] = led_value[i];
// den Abstand zwischen Ausgangs- und Zielwert ermitteln,
// mit 128 multiplizieren und durch die Schrittzahl dividieren
for (i = 0; i < LEDs; i++)
	{
	delta[i] = led_value[i] - ausgangswert[i]; // der Ausgangswert wurde beim letzten sort_byte() gespeichert
	delta[i] *= 128;
	delta[i] /= schrittzahl;
	zwischenwert[i] = (uint16_t) (ausgangswert[i] * 128);
	}

while(schrittzahl)								// fr jeden Zwischenschritt die Helligkeitswert interpolieren
	{
	for (i = 0; i < LEDs; i++)  				// Nchsten Zwischenwert ermitteln
		{
		zwischenwert[i] += delta[i]; 			// Zwischenwert beibehalten
		led_value[i] = (uint8_t) (zwischenwert[i] / 128); // und fr Ausgabe durch 128 dividieren
		}

	prep_values();									// hier wird pwm_update auf TRUE gesetzt.
														// am Beginn des nchsten Taktes geht pwm_update auf FALSE

	pwm_counter = pwm_wait;						// Warteschleife fr Verzgerung beim Faden
	while(pwm_counter);							// wartet, bis der Counter vom PWM-Start auf 0 decrementiert ist

	schrittzahl--;
	}
														// als letzten Wert den exakten Zielwert bergeben
														// dadurch werden unvermeidliche Rundungsfehler ausgeglichen
for (i = 0; i < LEDs; i++) led_value[i] = zielwert[i];
prep_values();										// ausgeben

pwm_busy = FALSE;									// via TWI kann abgefragt werden, ob ein bergang in Arbeit ist
}

// ----------------------------------------------------------------------------
// den Buffer in aufsteigender Folge sortieren
// dabei auch die zugehrigen Bitcodes mitsortieren
// so dass Helligkeitswert und Bitcode den selben Indexwert haben
// benutzt die globalen Variablen
// - led_value[LEDs]		// bermittelter 1-Byte-Wert fr die Helligkeit
// - led_bitcode[LEDs]	// das zu schaltende Port-Bit fr den zugeh. Kanal

void prep_values(void)
{
uint8_t ready = FALSE;
uint8_t i, j, k, t, cnt;
uint16_t tt;

// zum berblenden muss der letzte Helligkeitswert als Ausgangswert gesichert werden
for (i = 0; i < LEDs; i++)	ausgangswert[i] = led_value[i];

// die Bits fr die Portpins der jeweiligen LEDs setzen
for (i = 0; i < LEDs; i++) led_bitcode[i] = portbit[i];

// zustzlich noch die ein weiteres Byte anhngen, die beim Sortieren garantiert
// das Schlusslicht bilden - und alle Leds abschalten.
//led_bitcode[LEDs] = led_bitcode[LEDs + 1] = 0;
//led_value[LEDs]   = led_value[LEDs+1]     = 255;
led_bitcode[LEDs] = 0;
led_value[LEDs]   = 255;

cnt = (LEDs + 1);  // es ist noch ein '255' fr den letzten OCR1A-Wert angehngt worden

// den Buffer in aufsteigender Folge sortieren
// dabei auch die zugehrigen Bitcodes mitsortieren
// so dass Helligkeitswert und Bitcode den selben Indexwert haben
// die Sortierung erfolgt durch Vertauschen:
// immer wenn das n + 1. Element kleiner ist, werden n und (n + 1) vertauscht.
while (ready == FALSE)
	{
	ready = TRUE;

	for (i = 0; i < (cnt - 1); i++)
		{
		j = i + 1;
		if (led_value[i] > led_value[j])		// wenn die Werte gleich sind
			{											// dann vertauschen
			t = led_value[i];						// Helligkeitswerte
			led_value[i] = led_value[j];
			led_value[j] = t;

			t = led_bitcode[i];					// zugehrige Portbits
			led_bitcode[i] = led_bitcode[j];
			led_bitcode[j] = t;

			ready = FALSE;							// Abbruchbedingung fr Sortierung
			}
		}
	}

// zeitgleich stattfindende Schaltvorgnge finden,
// die BitCodes fr gleichzeitigen Schaltvorgnge zusammenfhren,
// die nachfolgenden Daten vorrcken lassen,
// nach hinten mit 255 auffllen
// den Zhler cnt fr die Gesamtzahl der Eintrge decrementieren.
//
// Zu beachten ist, dass es nicht ausreicht, die die vorgegebenen Helligkeitswerte
// auf Gleichheit zu testen (weil dann der Schaltzeitpunkt definitv identisch ist).
// Bedingt durch die bersetzungstabelle in OCR1A-Werte und deren sehr langsamen Anstieg
// am Anfang der Tabelle, knnen trotz unterschiedlicher Helligkeitswerte dennoch
// gleich OCR1A-Werte entstehen.
// Auch hier mssen die Werte bei Gleichheit zusammengelegt werden.
// Das Problem beschrnkt sich aber auf den "unteren" Rand der bersetzungstabelle.
i = 0;
while (i < cnt - 1)								// jedes Element mit seinem Nachfolger vergleichen
	{
// Beim Vergleich der Schaltzeitpunkte (OCRA-Werte) reicht es nicht aus, die bermittelten 8-Bit
// Helligkeitswerte zu vergleichen. Gerade am unteren Tabellenende sind die OCRA-Werte mehrerer
// Helligkeitswerte gleich. 
// Daher wird gleich ein direkter Vergleich der 16-Bit-OCRA-Werte durchgefhrt
		
	if (pgm_read_word(&ocra_tab[ led_value[i] ]) == pgm_read_word(&ocra_tab[ led_value[i + 1] ]))
		{
//		led_bitcode[i] |= led_bitcode[i+1]; // bitcodes addieren 
		led_bitcode[i] += led_bitcode[i+1]; // bitcodes addieren 

		for (j = i + 1; j < cnt - 1; j++)   // alle Nachfolgender nach vorne aufrcken lassen
			{
			led_value[j]   = led_value[j + 1];
			led_bitcode[j] = led_bitcode[j + 1];
			}
		cnt--;                     			// jetzt haben wir ein Element weniger
		led_value[cnt] = 255;					// hinten freigewordene mit '255' auffllen
		led_bitcode[cnt] = 0;					// hinten freigewordene mit '0' auffllen
		}

	else i++;										// wenn Ungleich, dann zum nchsten Element
	}

// Um die Beschickung des Timers1 ohne weitere Abfragen komplett ber die Arrays steuern zu knnen,
// muss ein Element an der Startposition vorhanden sein.
// Wenn keine der LEDs komplett ausgeschaltet ist, dann muss ein Element fr den PWM-Start
// angelegt werden.

if (led_value[0] > 0)      					// wenn das erste Element nicht 0 ist, dann sind alle LEDs
	{                    						// beim Start eingeschaltet, dazu einen bitcode einfgen
	i = cnt++;           						// es gibt dann 1 Element mehr im Array
	while (i )            						// dazu alles um eine Stelle nach hinten schieben
		{
		led_value[i]   = led_value[i - 1];
		led_bitcode[i] = led_bitcode[i- 1];
		i--;
		}
	led_value[0]   = 0;       					// freigewordenes 1.Element erhlt den Wert '0' fr Start
	led_bitcode[0] = LED_PINS; 				// der Bitcode fr den Start: alle LEDs an
	}

else													// mindestens eine LED hat den Helligkeitswert 0
	{
	led_bitcode[0] = ~led_bitcode[0];		// dann wird/werden sie erst garnicht eingeschaltet.
	}

// hier wird das logische Undieren vorweggenommen,
// so dass die ISR nur noch den bit_code auf den Port schreiben muss
// der Startwert ist bereits bekannt, also ab 2. Element wird immer mit dem Vorgnger Undiert
for (i = 1; i < cnt; i++)
	{
//	led_bitcode[i] = ( led_bitcode[i-1] & (~(led_bitcode[i])) );
	led_bitcode[i] = ( led_bitcode[i-1] - led_bitcode[i] );// in diesem Fall kann auch einfach subtrahiert werden
	}
														// hier werden die OCR1A-Werte und die Portbits seriell in
														// der spter bentigten Reihenfolge in ein Array geschrieben.
														// Das Array ist gedanklich in zwei Hlften geteilt:
														// whrend in die eine Hlfte neue Daten geschrieben werden knnen,
														// kann die ISR aus der anderen Hlfte die aktuellen Werte auslesen.
														// Nur der Index entscheidet, wer welche Hlfte benutzt.
if (isr_idx < 32)									// wenn der Index < 32, dann die oberen Variablen benutzten 
	{													// Die ISR nutzen als Zhlervariable das Register isr_idx.
	i = isr_start_idx = 32; 					// Der Startindex fr die neuen Werte (obere Arrayhlfte)
	}
	
else		
	{
	i = isr_start_idx = 0; 						// Der Startindex fr die neuen Werte (untere Arrayhlfte)
	}

j = k = 0;											

while(k < (LEDs))
	{
	tt = pgm_read_word(&ocra_tab[ led_value[k + 1] ]);
	pwm_buffer[(j++) + i] = (uint8_t) (tt >> 8);
	pwm_buffer[(j++) + i] = (uint8_t) (tt);
	
#ifdef HIGHSIDE
	pwm_buffer[(j++) + i] = ~led_bitcode[k++] ;	// wenn zum EINgeschalten der Pin auf low gezogen werden muss
#else
	pwm_buffer[(j++) + i] =  led_bitcode[k++] ;	// wenn zum EINgeschalten der Pin auf high gezogen werden muss
#endif
	}

pwm_update = TRUE;								// die Werte werden beim nchsten PWM-Start bernommen
														// und das Flag wird dabei zurckgesetzt.
while (pwm_update);								// Warten, bis die Daten bernommen worden sind.
}

// --- [ eof ] ----------------------------------------------------------------